'use strict';

const hljs = require('highlight.js');
const stripIndent = require('strip-indent');
const alias = require('../highlight_alias.json');

const reg_line_str = /<span class="hljs-.*">(.*?)<\/span>/;

function highlightUtil(str, options = {}) {

  if (typeof str !== 'string') throw new TypeError('str must be a string!');
  str = stripIndent(str);

  const {
    firstLine = 1,
    caption,
    mark = [],
    tab
  } = options;

  hljs.configure({ classPrefix: 'hljs-' });

  const data = highlight(str, options);
  const lines = data.value.split('\n');
  let content = '';

  for (let i = 0, len = lines.length; i < len; i++) {
    let line = lines[i];
    if (tab) line = replaceTabs(line, tab);
    content += formatLine(line, Number(firstLine) + i, mark, options);
  }

  let codeCaption = '';

  if (caption) {
    codeCaption = `<div class="codeBlockTitle">${caption}</div>`
  }

  let result = `<div class="codeBlockContainer">${codeCaption}`
  result += `<div class="codeBlockContent">`
  result += `<pre class="highlight">`
  result += `<code class="${data.language ? ` language-${data.language}` : ''} codeBlockLines">${content}`
  result += `</code></pre></div></div>`
  return result

}

function formatLine(line, lineno, marked, options = {}) {
  line = line.trim() === '' ? '</br>' : line;

  // 将 line 匹配 reg_line_str 正则表达式
  // 获取匹配的内容，判断是否为空
  const match = line.match(reg_line_str);
  if (match) {
    const content = match[1];
    if (content.trim() === '') {
      line = line.replace(reg_line_str, '</br>');
    }
  }

  let res = '<span class="line';
  if (marked.includes(lineno)) {
    // Handle marked lines.
    res += ` marked">${line}</span>`;
  } else {
    res += `">${line}</span>`;
  }

  res += '\n';
  return res;
}

function replaceTabs(str, tab) {
  return str.replace(/^\t+/, match => {
    let result = '';

    for (let i = 0, len = match.length; i < len; i++) {
      result += tab;
    }

    return result;
  });
}

function highlight(str, options) {
  let { lang } = options;
  const { autoDetect = false } = options;

  if (lang) {
    lang = lang.toLowerCase();
  } else if (autoDetect) {
    const result = hljs.highlightAuto(str);
    return closeTags(result);
  }

  if (!lang || !alias.aliases[lang]) {
    lang = 'plaintext';
  }

  const res = hljs.highlight(str, {
    language: lang,
    ignoreIllegals: true
  });

  return closeTags(res);
}

// https://github.com/hexojs/hexo-util/issues/10
function closeTags(res) {
  const tokenStack = [];

  res.value = res.value.split('\n').map(line => {
    const prepend = tokenStack.map(token => `<span class="${token}">`).join('');
    const matches = line.matchAll(/(<span class="(.*?)">|<\/span>)/g);
    for (const match of matches) {
      if (match[0] === '</span>') tokenStack.shift();
      else tokenStack.unshift(match[2]);
    }
    const append = '</span>'.repeat(tokenStack.length);
    return prepend + line + append;
  }).join('\n');
  return res;
}

module.exports = highlightUtil;
